;; All hail the RECV form
(define-runtime-macro (recv . clauses)
  (let ((msg  (gensym 'msg))   ;; the current mailbox message
        (loop (gensym 'loop))) ;; the mailbox seeking loop
    
    ;; check the last clause to see if it's a timeout
    (let ((sesualc (reverse clauses)))
      (if (and (pair? (car sesualc))
               (eq? (caar sesualc) 'after))

          (let ((clauses (reverse (cdr sesualc)))
                ;; the code to compute the timeout
                (init (cadar sesualc))
                ;; the variable holding the timeout
                (timeout (gensym 'timeout))
                ;; the code to be executed on a timeout
                (on-timeout (cddar sesualc))
                ;; the timeout exception-handler to the whole match
                (e (gensym 'e)))

            ;; RECV code when there is a timeout
            `(let ((,timeout ,init))
               (with-exception-catcher
                (lambda (,e)
                  (if (mailbox-receive-timeout-exception? ,e)
                      (begin
                        (thread-mailbox-rewind)
                        ,@on-timeout)
                      (raise ,e)))
                (lambda ()
                  (let ,loop ((,msg (thread-mailbox-next ,timeout)))
                       (match/action
                        (thread-mailbox-extract-and-rewind)
                        (,loop 
                         (thread-mailbox-next ,timeout))
                        ,msg
                        ;; extra clause to handle system events
                        (event 
                         (where (termite-exception? event))
                         (handle-exception-message event))
                        ;; the user clauses
                        ,@clauses))))))

          ;; RECV code when there is no timeout
          `(let ,loop ((,msg (thread-mailbox-next)))
                (match/action
                 (thread-mailbox-extract-and-rewind)
                 (,loop
                  (thread-mailbox-next))
                 ,msg
                 ;; extra clause to handle system events
		 (event 
		  (where (termite-exception? event))
		  (handle-exception-message event))
                 ;; the user clauses
                 ,@clauses))))))

(define-runtime-macro (make-process . code)
  `(spawn 
    (lambda ()
      ,@code)))

(define-runtime-macro (make-process/recv . code)
  `(make-process
    (let loop ()
      (recv 
        ((from tag ('process-migrate/proxy to))
          (migrate/proxy to)
    	  (! from (list tag (self))))
        ((from tag ('process-migrate to))
    	    (migrate-task to)
    	    (! from (list tag (self))))
        ((from tag 'process-clone)
    	(call/cc
    	    (lambda (clone)
    		(! from (list tag clone)))))
    	(('process-update cont)
    	 (cont #t))
	,@code)
      (loop))))

(define-runtime-macro (make-process/recv/locals locals . code)
  `(make-process
    (let loop () 
        (let* (,@locals)
         (recv 
        ((from tag ('process-migrate/proxy to))
          (migrate/proxy to)
    	  (! from (list tag (self))))
        ((from tag ('process-migrate to))
    	    (migrate-task to)
    	    (! from (list tag (self))))
        ((from tag 'process-clone)
    	(call/cc
    	    (lambda (clone)
    		(! from (list tag clone)))))
    	(('process-update cont)
    	 (cont #t))
      ,@code)
      (loop)))))

(define-runtime-macro (define/process name . code)
  `(define ,name
     (make-process ,@code)))

(define-runtime-macro (define/process/recv name . code)
  `(define ,name
     (make-process/recv ,@code)))

(define-runtime-macro (make-process/link . code)
  `(spawn-link 
    (lambda ()
      ,@code)))

(define-runtime-macro (make-process/link/recv . code)
  `(make-process/link
    (let loop ()
      (recv 
        ((from tag ('process-migrate/proxy to))
          (migrate/proxy to)
    	  (! from (list tag (self))))
        ((from tag ('process-migrate to))
    	    (migrate-task to)
    	    (! from (list tag (self))))
        ((from tag 'process-clone)
    	(call/cc
    	    (lambda (clone)
    		(! from (list tag clone)))))
    	(('process-update cont)
    	 (cont #t))
      ,@code)
      (loop))))

(define-runtime-macro (make-process/link/recv/locals locals . code)
  `(make-process/link
    (let loop () 
        (let* (,@locals)
         (recv 
        ((from tag ('process-migrate/proxy to))
          (migrate/proxy to)
    	  (! from (list tag (self))))
        ((from tag ('process-migrate to))
    	    (migrate-task to)
    	    (! from (list tag (self))))
        ((from tag 'process-clone)
    	(call/cc
    	    (lambda (clone)
    		(! from (list tag clone)))))
    	(('process-update cont)
    	 (cont #t))
      ,@code)
      (loop)))))

(define-runtime-macro (define/process/link name . code)
  `(define ,name
     (make-process/link ,@code)))

(define-runtime-macro (define/process/link/recv name . code)
  `(define ,name
     (make-process/link/recv ,@code)))
